package rpss15

trait Future[+T] { self =>
  def onComplete(callback: Try[T] => Unit)

  def map[U](f: T => U) = new Future[U] {
    def onComplete(callback: Try[U] => Unit) =
      self onComplete (t => callback(t.map(f)))
  }

  def flatMap[U](f: T => Future[U]) = new Future[U] {
    def onComplete(callback: Try[U] => Unit) =
      self onComplete { _.map(f) match {
        case Success(fu) => fu.onComplete(callback)
        case Failure(e)  => callback(Failure(e))
    } }
  }

  def filter(p: T => Boolean) =
    map { t => if (!p(t)) throw new NoSuchElementException; t }
}

object Future {
  def apply[T](f: => T) = {
    val handlers = collection.mutable.Buffer.empty[Try[T] => Unit]
    var result: Option[Try[T]] = None

    val runnable = new Runnable {
      def run = {
        val r = Try(f)
        handlers.synchronized {
          result = Some(r)
          handlers.foreach(_(r))
        }
      }
    }

    (new Thread(runnable)).start()

    new Future[T] {
      def onComplete(f: Try[T] => Unit) = handlers.synchronized {
        result match {
          case None => handlers += f
          case Some(r) => f(r)
        }
      }
    }
  }
}
